/*
 *  Copyright 2007, 2008, 2009 eadmaster <spamhereplease@fastwebnet.it>
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#include <SDL.h>
#include "SDL_Utils.h"


void SDL_SetPixel( SDL_Surface *surface, int x, int y, Uint32 pixel )
//code from http://www.libsdl.org/cgi/docwiki.cgi/Pixel_Access
{
    if( (x>=(surface->w)) || (y>=(surface->h)) ) return; // bounds check

    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}


Uint32 SDL_GetPixel( SDL_Surface *surface, int x, int y )
//code from http://www.libsdl.org/cgi/docwiki.cgi/Pixel_Access
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}


void SDL_ScaleSurface( SDL_Surface *src, SDL_Surface *dst, Uint16 NewWidth, Uint16 NewHeight )
// code from http://www.sdltutorials.com/sdl-scale-surface/
{
    if(!src || !dst || !NewWidth || !NewHeight)
        return;

   // SDL_Surface *Surface=src;
    //SDL_Surface *_ret = dst;

    double  _stretch_factor_x = ((double)(NewWidth) / (double)(src->w)),
        _stretch_factor_y = ((double)(NewHeight) / (double)(src->h));

    Sint32 y,x,o_y,o_x;
    for(y = 0; y < src->h; y++)
        for(x = 0; x < src->w; x++)
            for(o_y = 0; o_y < _stretch_factor_y; ++o_y)
                for(o_x = 0; o_x < _stretch_factor_x; ++o_x)
                    SDL_SetPixel(dst, (Sint32)(_stretch_factor_x * x) + o_x, (Sint32)(_stretch_factor_y * y) + o_y, SDL_GetPixel(src, x, y));

}

SDL_Surface* SDL_CreateSurface(Uint16 Width, Uint16 Height)
{
    return SDL_CreateRGBSurface( SDL_HWSURFACE,
                                 Width,
                                 Height,
                                 SDL_GetVideoSurface()->format->BitsPerPixel,
                                 SDL_GetVideoSurface()->format->Rmask,
                                 SDL_GetVideoSurface()->format->Gmask,
                                 SDL_GetVideoSurface()->format->Bmask,
                                 SDL_GetVideoSurface()->format->Amask );
    // no need to call SDL_DisplayFormat?
}

SDL_Surface* SDL_TrimSurface(SDL_Surface* s, SDL_Rect r)
{
    SDL_Surface* output = SDL_CreateSurface(r.w, r.h);
    SDL_BlitSurface(s, &r, output, NULL);
    return(output);
}

SDL_Surface* SDL_MergeSurfaces(SDL_Surface* s[])
{
    Uint16 outputW = 0;
    Uint16 outputH = 0;

    int ssize = sizeof(s) / sizeof(SDL_Surface*);

    int i;
    for(i=0; i<ssize; i++)
    {
        outputW += (s[i]->w); // sum up all weights

        if((s[i]->h) > outputH) // find max height
            outputH = (s[i]->h);
    }

    SDL_Surface* output = SDL_CreateSurface(outputW, outputH);

    SDL_Rect dstrect = {0,0,0,0};
    for(i=0; i<ssize; i++)
    {
        dstrect.h = ( outputH - (s[i]->h) ) / 2;

        SDL_Rect dstrectCopy = dstrect;
        SDL_BlitSurface(s[i], NULL, output, &dstrectCopy);

        dstrect.w += (s[i]->w);
    }

    return(output);
}

/* Basic unantialiased Bresenham line algorithm
 code from http://lists.libsdl.org/pipermail/sdl-libsdl.org/1999-November/003589.html
 see also http://en.wikipedia.org/wiki/Bresenham_algorithm
*/
#define SGN(x) ((x)>0 ? 1 : ((x)==0 ? 0:(-1)))
#define ABS(x) ((x)>0 ? (x) : (-x))
void SDL_DrawLine( SDL_Surface* s, Uint32 x1, Uint32 y1, Uint32 x2, Uint32 y2, Uint32 color)
{
  int lg_delta, sh_delta, cycle, lg_step, sh_step;

  lg_delta = x2 - x1;
  sh_delta = y2 - y1;
  lg_step = SGN(lg_delta);
  lg_delta = ABS(lg_delta);
  sh_step = SGN(sh_delta);
  sh_delta = ABS(sh_delta);
  if (sh_delta < lg_delta) {
    cycle = lg_delta >> 1;
    while (x1 != x2) {
      SDL_SetPixel(s, x1, y1, color);
      cycle += sh_delta;
      if (cycle > lg_delta) {
	cycle -= lg_delta;
	y1 += sh_step;
      }
      x1 += lg_step;
    }
    SDL_SetPixel(s, x1, y1, color);
  }
  cycle = sh_delta >> 1;
  while (y1 != y2) {
    SDL_SetPixel(s, x1, y1, color);
    cycle += lg_delta;
    if (cycle > sh_delta) {
      cycle -= sh_delta;
      x1 += lg_step;
    }
    y1 += sh_step;
  }
  SDL_SetPixel(s, x1, y1, color);
}
