/*
 *  libmSDL
 *  Copyright 2007, 2008, 2009 eadmaster <spamhereplease@fastwebnet.it>
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#ifndef _MSDL_H_
#define _MSDL_H_

// SDL includes
#include <SDL.h>

// standard C++ includes
#include <iostream>
#include <set>
#include <iterator>
//#include <string>
//#include <algorithm>
//#include <vector>
//#include <functional>
//#include <cmath>
#include "pmath.h"
#include "SDL_Utils.h"

// mSDL version
#define _MSDL "0.9a"


namespace mSDL
{

/* PAlib-styled signed and unsigned types */
typedef Sint8	s8;  //!< 8-bit signed integer
typedef Uint8	u8;  //!< 8-bit unsigned integer
typedef Sint16  s16; //!< 16-bit signed integer
typedef Uint16  u16; //!< 16-bit unsigned integer
typedef Sint32	s32; //!< 32-bit signed integer
typedef Uint32  u32; //!< 32-bit unsigned integer

/* remove name decorating from some SDL-specific types */
typedef SDL_Rect Rect; //!< see http://www.libsdl.org/cgi/docwiki.cgi/SDL_Rect
typedef SDL_Rect Point;
typedef SDL_Surface Surface; //!< see http://www.libsdl.org/cgi/docwiki.cgi/SDL_Surface
typedef SDL_Color Color; //!< see http://www.libsdl.org/cgi/docwiki.cgi/SDL_Color

#include "mSDL_Consts.hpp"
#include "mSDL_Colors.hpp"


// GRAPHIC SYSTEM
    /*! \brief Initialize (m)SDL. Need to be invoked before using any mSDL module.
     * Initialize SDL subsystems used by mSDL (video, audio,...).
     * Abort execution if it fails.
     * If the logical screen size is not equal to the real screen the surface is automatically scaled appropriately (this is useful for cross-platform porting).
     */
    void Init( const u16 LogicalScreenWidth = DEFAULT_SCREEN_WIDTH, const u16 LogicalScreenHeight = DEFAULT_SCREEN_HEIGHT );

    extern SDL_Surface* Screen; //!< The logical screen surface. User blits should be done to this surface instead of the one returned by SDL_GetVideoSurface.
    // TODO? rename as Video/Display

    //! Save a screenshot into a BMP file.
    inline void Screenshot( const std::string Filename )
    {
        if( SDL_SaveBMP( Screen, Filename.c_str() ) != 0 )
            std::cerr << "mSDL::Screenshot(" << Filename << "): " << SDL_GetError() << std::endl;
    }

    u8 GetFPS( void ); //!< get rendered Frames-Per-Second.

    #include "mSDL_GFX.hpp"
    #include "mSDL_Screen.hpp"
    #include "mSDL_Sprite.hpp"
    #include "mSDL_BG.hpp"
    // TODO: mSDL_Font ? -> Banner extends Font
    #include "mSDL_Banner.hpp"


// INPUT SYSTEM
    struct ButtonStatus
    {
        bool Held; //!< true in ANY instant the button is held down
        bool Newpress; //!< true only in the first instant the button is pressed
        bool Released; //!< true only in the first instant the button is released
    };

    //TODO: #include "mSDL_Input.hpp" -> rimpiazza mSDL_Key/Mouse
    // REMOVE?
    #include "mSDL_Key.hpp"
    #if TARGET_HAS_KEYBOARD
        extern KeyHandler Key; //!< default instance from which to read keyboard buttons state
    #endif

    #include "mSDL_Mouse.hpp"
    #if TARGET_HAS_MOUSE
        extern MouseHandler Mouse; //!< default instance from which to read mouse position and buttons state
        extern MouseHandler& Stylus;
    #endif

    #if TARGET_HAS_JOY
        //TODO: Pad/Joy
    #endif


// AUDIO SYSTEM
    #include "mSDL_Music.hpp"
    #include "mSDL_Sound.hpp"

// GENERAL
    //! sleep for the specified number of milliseconds.
    inline void Sleep( const u32 ms )
    {
        SDL_Delay( ms );
    }

    //! Very important function: it draws the screen and read user input. Must be invoked only once for every program cycle. (Similar to PA_WaitForVBL.)
    inline void WaitForVBL( void )
    {
        ScreenObject::DrawAll();
        Key.UpdateStatus();
        Mouse.UpdateStatus();
    }

}

#endif /* _MSDL_H_ */
