/*
 *  libmSDL
 *  Copyright 2007, 2008, 2009 eadmaster <spamhereplease@fastwebnet.it>
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

/*! \file mSDL_GFX.hpp
    \class GFX
    \brief A graphic object. It handles surfaces manipulation and animation.
*/

#ifdef _HAS_SDL_IMAGE
    #include <SDL_image.h>
#endif

#ifdef _HAS_OPENGL
    #include <SDL_opengl.h>
#endif


class GFX
{
    protected:

        SDL_Surface* Surface;
        SDL_Color Transparent;
        u8 Alpha;

        // zooming/rotation handling
            bool RotZoomEnabled;
            SDL_Surface *OriginalSurface; // original surface before rotation or zooming, used for reversing transformations
            //float ZoomX;
            //float ZoomY;
            //float RotationAngle;

        // frame animation handling:
            u16 W, H; // width and height of a frame (in pixels)
            u8 NoFrames; // total number of frames in the spritesheet
            u16 Frame; // currently showing frame number (count starts from 0)
            bool Anim; // true if playing animation (cicling through designed frames)
            u32 LastFrameTime; // last (SDL) time there was a frame change
            void UpdateFrame( void ); // choose correct frame according to current animation setups
            SDL_Rect GetFrameRect() const; // compute current frame

        // animation setups:
            //u8 FirstFrame; // first frame in current playing animation (default: 0)
            //u8 LastFrame; // last frame in current playing animation (default: NoFrames)
            u8 FPS; // frames-per-second shown (default: 60)
            //bool AnimLoop; //!< true if looping frames circularly (next to last frame is first frame)
            //bool AnimLeftRight; //!< true if looping frames Left-Right
            s8 AnimNoLoops; // number of remaining loops before to stop. (0 = stop now, default: -1 = infinite)

	private:

		void Init() // common initializer
		{
			Surface = NULL;
			Transparent = NONE;
			Alpha = SDL_ALPHA_OPAQUE;

			RotZoomEnabled = false;
			OriginalSurface = NULL;

			W = H = 0;
			NoFrames = 1; // or 0?
			Frame = 0;
			Anim = false;

			//FirstFrame = 0;
			//LastFrame = NoFrames;
			FPS = 60;
			AnimNoLoops = -1;
		}

        void SetDisplayFormat() // optimize the Surface for fast blitting
        {
            if( !Surface ) return;
            SDL_Surface* t = Surface; // create a copy of previous surface pointer for later freeing

            /*if(GetAlpha() != SDL_ALPHA_OPAQUE) // if alpha is set
            {
                Surface = SDL_DisplayFormatAlpha( t );
                // MEMO: set alpha=0? -> use Alpha private member?
                //std::cout << std::endl << ((int) GetAlpha()) << std::endl;
            }
            else*/
            {
                Surface = SDL_DisplayFormat( t );
            }

            if( !Surface )
            {
                std::cerr << "mSDL::GFX::SetDisplayFormat: " << SDL_GetError() << std::endl;
                exit( 1 );
            }

            SDL_FreeSurface( t ); // free the previous surface
        }


    public:

        //! default constructor:
        //! create an empty/invisible surface
        GFX()
        {
            Init();
        }

        //! most-used constructor:
        //! load a surface from a graphic file
        //! (only BMPs are supported, use SDL_image and the second constructor
        //! to load other image formats)
        GFX( const std::string Filename,
             const SDL_Color InitTransparent = NONE,
             const u16 InitNoFrames = 1 )
        {
            Init();

            // load the surface from the specified filename
#ifdef _HAS_SDL_IMAGE
            Surface = IMG_Load( Filename.c_str() );
#else
            Surface = SDL_LoadBMP( Filename.c_str() );
#endif
            if( !Surface )
            {
                // loading error
                std::cerr << "mSDL::GFX(...): " << SDL_GetError() << std::endl;
                return; // leave the sprite invisible
                //ALTERNATIVE BEHAVIOR: exit( 1 ); // abort execution
            }

            SetTrans( InitTransparent ); // also SetDisplayFormat()
            SetNoFrames( InitNoFrames );
        }

        //! alternative constructor no.1:
        //! uses a given SDL_Surface
        //! NOTE: InitSurface is copied and can be safely fred after invocation.
        GFX( SDL_Surface *InitSurface,
             const SDL_Color InitTransparent = NONE,
             const u16 InitNoFrames = 1 )
        {
            Init();
            SetSurface( InitSurface ); // also SetDisplayFormat()
            SetTrans( InitTransparent ); // also SetDisplayFormat()
            SetNoFrames( InitNoFrames );
        }


        //! alternative constructor no.2:
        //! create an empty surface of given dimensions (default is screen size)
        //! and fills it with a solid color
        GFX( SDL_Color InitColor,
             u16 Width = Screen->w,
             u16 Height = Screen->h )
        {
            Init();

            Surface = SDL_CreateRGBSurface( SDL_HWSURFACE, Width, Height, Screen->format->BitsPerPixel,
                                            Screen->format->Rmask, Screen->format->Gmask, Screen->format->Bmask, Screen->format->Amask );
            SDL_FillRect( Surface, NULL, SDL_MapRGB( Surface->format, InitColor.r, InitColor.g, InitColor.b ) );

            SetDisplayFormat();
            SetNoFrames( 1 );
        }

        // TODO? operator+ -> statically combine/merge one GFX into another
        // TODO? operator>> -> combine one GFX into another (inplace)
        // TODO? operator<< -> return debug info about a GFX object

        GFX Clone( void ) const
        {
            GFX c = *this; // clone this object using copy constructor

            c.Surface = SDL_CloneSurface( this->Surface ); // also clone this object surface

            if(RotZoomEnabled)
                c.OriginalSurface = SDL_CloneSurface( this->OriginalSurface );

            return c;
        }

        //! saves this graphical surface into a BMP file.
        void Save( const std::string Filename ) const
        {
            if( SDL_SaveBMP( Surface, Filename.c_str() ) != 0 )
                std::cerr << "mSDL::GFX::Save(" << Filename << "): " << SDL_GetError() << std::endl;
        }

/* REMOVE?
        void Clear( void ) //!< Clear the GFX object. All ScreenObjects using this GFX will appear invisible. NOTE: It does not affect the "Visible" field and it can't be undone (surface memory is fred).
        {
            StopAnim();
            if( Surface ) SDL_FreeSurface( Surface );
            Init();
        }
*/
        SDL_Surface* GetSurface() const { return Surface; } //!< return the associated SDL Surface for direct manipulation
        void SetSurface( SDL_Surface* NewSurface ) //! NewSurface is copied and can be safely fred after invocation.
        {
            //NO? SDL_FreeSurface( Surface );
            Surface = NewSurface;
            SetDisplayFormat();
        }

        u16 GetWidth() const { return(W); }
        u16 GetHeight() const { return(H); }

        u8 GetNoFrames() { return(NoFrames); }
        void SetNoFrames( const u16 NewNoFrames = 1 )
        {
        	StopAnim(); // must stop any animation

        	NoFrames = NewNoFrames; // change current number of frames

        	if(!Surface) return;

        	// set the real Width and Height of the surface depending on the number of frames
            W = Surface->w;
            H = Surface->h;

            // if Surface is a spritesheet with multiple frames
            // must reduce GFX' width and height
            if( NoFrames > 1 )
            {
                if( (Surface->w) > (Surface->h) )
                {
                    // spritesheet is horizontal
                    W = ( Surface->w / NoFrames );
                }
                else
                {
                    // spritesheet is vertical
                    H = ( Surface->h / NoFrames );
                }
            }
        }

//TODO: rename as "Get/SetTransparentColor"?
        SDL_Color GetTrans( void ) const { return Transparent; }
        void SetTrans( const SDL_Color NewTransparent = NONE ) //!< set the transparent color. Invoked with no arguments will remove any previously set transparency.
        {
            if( !Surface ) return;

            Transparent = NewTransparent;

            if( Transparent.unused == 1 )
                SDL_SetColorKey( Surface, 0, 0 );
            else
                SDL_SetColorKey( Surface, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB( Surface->format, Transparent.r, Transparent.g, Transparent.b ) );

            SetDisplayFormat();
        }

        void SetAlpha( const u8 NewAlpha = SDL_ALPHA_OPAQUE ) //!< set the alpha blending. If invoked with no args reset to opaque.
        {
            if( !Surface ) return;

            Alpha = NewAlpha;

            SDL_SetAlpha(Surface, SDL_SRCALPHA | SDL_RLEACCEL, Alpha);

            SetDisplayFormat();
        }

        u8 GetAlpha() const
        {
            //return(Surface->format->alpha);
            return(Alpha);
        }


        // ********** animation handling methods *****

        u16 GetFrame() const { return Frame; } //!< get current frame number (count starts from 0)
        void SetFrame( const u16 NewFrame = 0 ){ if( NewFrame < NoFrames ) Frame = NewFrame; } //!< set current frame number (count starts from 0, must be < NoFrames)

        bool isAnim() const { return Anim; } //!< return true if playing animation

        void StartAnim() { Anim = true; LastFrameTime = SDL_GetTicks(); } //!< start sprite auto-animation
        void PauseAnim() { Anim = false; }
        void StopAnim() { Anim = false; Frame = 0; }

        u8 GetFPS() { return(FPS); }
        //TODO: Get/SetFirstFrame
        //TODO: Get/SetLastFrame
        //TODO: Get/SetFPS(u8 NewFPS); //!< set desired FPS rate limit


        // ********** pixel-manipulation methods *****
        SDL_Color GetPixel( const u16 X, const u16 Y ) const
		{
			SDL_Color t;
			SDL_GetRGB( SDL_GetPixel( Surface, X, Y ), Surface->format, &t.r, &t.g, &t.b );
			return t;
		}

        void SetPixel( const u16 X, const u16 Y, const SDL_Color Color = BLACK )
		{
		    SDL_SetPixel( Surface, X, Y, SDL_MapRGB( Surface->format, Color.r, Color.g, Color.b ) );
        }

        // NO? void DrawLine( u16 StartX, u16 StartY, u16 EndX, u16 EndY )
        void DrawLine( const u16 X1, const u16 Y1, const u16 X2, const u16 Y2,
                       SDL_Color EdgeColor = BLACK, u8 Thick = 1 )
        {
            Point Start = { X1, Y1 };
            Point End = { X2, Y2 };

            // checks if points are aligned
            if( Start.x == End.x )
            {
                if( Start.y == End.y )
                {
                    // need to draw a single point
                    SetPixel( Start.x, Start.y, EdgeColor );
                    // ALTERNATIVE: DrawCircle(Start.x, Start.y,Thick); // using thickness as radius
                    return;
                }

                // it is a vertical line
                Rect VLine = { Start.x, End.x, Thick, abs(Start.y - End.y) };
                DrawRect( VLine, EdgeColor );
                return;
            }
            else if( Start.y == End.y )
            {
                // it is an horizontal line
                Rect HLine = { Start.x, End.x, abs(Start.x - End.x), Thick };
                DrawRect( HLine, EdgeColor );
                return;
            }
            else
            {
                //the line is oblique
                // TODO: add width support? http://www.codeguru.com/forum/showthread.php?t=330717
                SDL_DrawLine(Surface, Start.x, Start.y, End.x, End.y, SDL_MapRGB( Surface->format, EdgeColor.r, EdgeColor.g, EdgeColor.b ));
            }
        }

        void DrawRect( Rect r, SDL_Color FillColor = BLACK )
        {
            // 2 TEST: need to copy r?
            SDL_FillRect( Surface, &r, SDL_MapRGB( Surface->format, FillColor.r, FillColor.g, FillColor.b ) );
        }

        //TODO: void DrawCircle( Point Center, u8 Radius );

/* use GetSurface() + SDL_gfx or SDL_draw for more drawing functions... */

/* TODO: rivedere e completare

        void EnableRotZoom() //RIMUOVERE?
        {
            OriginalSurface = Surface; // backup original surface
            RotZoomEnabled = true;
        }
        void DisableRotZoom() //RIMUOVERE?
        {
            Surface = OriginalSurface; // restore original surface
            RotZoomEnabled = false;
            SetNoFrames( NoFrames );
        }

        void Scale( const u16 NewWidth = W, const u16 NewHeight = H )
        {
            if(!RotZoomEnabled) return;

            NewWidth *= NoFrames;
            NewHeight *= NoFrames;

#ifdef _HAS_OPENGL
                // perform zooming in hardware with opengl

                // TODO!
#else
                // perform scaling via software (slow)

                // create a new surface
                Surface = SDL_CreateRGBSurface( SDL_HWSURFACE,
                                               NewWidth,
                                               NewHeight,
                                               OriginalSurface->format->BitsPerPixel,
                                               OriginalSurface->format->Rmask,
                                               OriginalSurface->format->Gmask,
                                               OriginalSurface->format->Bmask,
                                               OriginalSurface->format->Amask );
                SDL_ScaleSurface(OriginalSurface, Surface, NewWidth, NewHeight);

                // reset transparency, noframes and alpha on new surface
                SetTrans( Transparent ); // also SetDisplayFormat()
                // TODO: must also re-set alpha
                //NOT NECESSARY? SetNoFrames( NoFrames );
#endif
        }
        void SetZoom( const float ZoomX = 1, const float ZoomY = 1 ) //! Just an alias to scale
        {
                Scale(GetWidth()*ZoomX, GetHeight()*ZoomY);
        }
        float GetZoomX() { return(ZoomX); }
        float GetZoomY() { return(ZoomY); }

        //TODO: void Rotate( float Angle ); -> SDL_RotateSurface
        {
            if(NoFrames>1)
            {
                // must rotate all frames separately!
            }

        }

        //TODO: void Flip( bool Direction = true ); // true = vertical, false = horizontal
*/

}; // END of class GFX
