/*
 *  libmSDL
 *  Copyright 2007, 2008, 2009 eadmaster <spamhereplease@fastwebnet.it>
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

/*! \file mSDL_Screen.hpp
    \class ScreenObject
    \brief A screen object. It handles objects positioning and drawing on screen.
*/

class ScreenObject : public GFX
{
    protected:

        bool Visible; //!< if false makes the object invisible. Can be useful in a lot of situations...

        s16 X, Y; //!< current screen position (pixel coords) for Sprites and Banners. For BGs indicate current scrolling position (the upper-left corner).
        s16 VX, VY; //!< velocities (in pixel/sec)
        u32 LastMoveTime; //!< last (SDL) time when position changed
        //s16 VA; //!< angle velocity
        //s16 A; //!< acceleration

        ScreenObject()
        {
            Visible = true;
            X = Y = VX = VY = 0;
            LastMoveTime = SDL_GetTicks();
            //Prio = InitPrio; Register(); -> put in derived classes
        }

        ~ScreenObject()
        {
            Visible = false;
            Unregister();
        }

        void Move( void ); //!< change object's position according to current velocities
        //OLD VERSION: void Move( void ) { SetXY( ( X + VX ), ( Y + VY ) ); }

        // void Rotate(); // rotate object according to current VA

        // priority system
            u16 Prio;
            /// @cond DEV
            struct CompareByPrio {
                bool operator()( const ScreenObject* a, const ScreenObject* b ) { return( (a->Prio) > (b->Prio) ); }
            };
            /// @endcond
            typedef std::multiset<ScreenObject*, CompareByPrio> MultisetSortedByPrio;
            static MultisetSortedByPrio RegisteredScreenObjects;
            MultisetSortedByPrio::const_iterator ROs_Pos; // store this object position in RegisteredScreenObjects' set
            void Register() { ROs_Pos = ( RegisteredScreenObjects.insert( this ) ); }
            void Unregister() { RegisteredScreenObjects.erase( ROs_Pos ); }


    public:

        bool isVisible() { return(Visible); }
        void SetVisible(bool b = true) { Visible=b; }
        void Show() { SetVisible(true); };
        void Hide() { SetVisible(false); };

        u16 GetPrio() const { return Prio; } //!< return the object priority
        virtual void SetPrio( const u16 NewPrio )
        {
            Unregister();
            Prio = NewPrio;
            Register();
        }  //!< change the object priority (the higher the most on top)

        s16 GetX() const { return ( X ); } //!< return the horizontal screen coordinate of the upper-left corner of the object
        s16 GetY() const { return ( Y ); } //!< return the vertical screen coordinate of the upper-left corner of the object
        s16 GetVX() const { return ( VX ); } //!< return the horizontal object velocity
        s16 GetVY() const { return ( VY ); } //!< return the vertical object velocity

        Point GetPosition( void ) const
        {
            SDL_Rect Position;
            Position.x = GetX();
            Position.y = GetY();
            return Position;
        }

        void SetX( const s16 NewX = 0 ) { X = NewX; }
        void SetY( const s16 NewY = 0 ) { Y = NewY; }
        virtual void SetXY( const s16 NewX = 0, const s16 NewY = 0 )
        {
            SetX( NewX );
            SetY( NewY );
        };

        void SetVX( const s16 NewVX = 0 ) //!< Set the horizontal speed (in pixel/sec)
        {
            if( NewVX != VX )
            {
                VX = NewVX;
                LastMoveTime = SDL_GetTicks();
            }
        }
        void SetVY( const s16 NewVY = 0 ) //!< Set the vertical speed (in pixel/sec)
        {
            if( NewVY != VY )
            {
                VY = NewVY;
                LastMoveTime = SDL_GetTicks();
            }
        }
        void SetVXY( const s16 NewVX = 0, const s16 NewVY = 0 )
        {
            SetVX( NewVX );
            SetVY( NewVY );
        }

        bool isMoving() const { return ( VX || VY ); }
        //bool isStatic() const { return !isMoving(); }


        //TODO: void ProcessAll( void (*CallbackFunction)(void) ); // process all registered screen objects with passed function's pointer
        // -> rimpiazza la drawAll
        //TODO: void ProcessAllSprites( void (*CallbackFunction)(void) ); // process all registered screen objects with passed function's pointer
        //TODO: void ProcessAllBanners( void (*CallbackFunction)(void) ); // process all registered screen objects with passed function's pointer
        //TODO: void ProcessAllBGs( void (*CallbackFunction)(void) ); // process all registered screen objects with passed function's pointer
        virtual void Draw( void ); // draw a single ScreenObjects over the Screen surface
        //REMOVE? virtual void Draw() = 0; // draw a single ScreenObjects over the Screen surface (pure virtual method, must be overloaded in derived classes)
        static void DrawAll( void ); // draw all registered ScreenObjects according to their priorities (invoking their virtual Draw methods)


        // TODO: test the followings:
        /*
        void SetXCenter( void ) { SetX( half( Screen->w ) - half( GetWidth() ) ); } //!< Center object horinzontally on the screen
        void SetYCenter( void ) { SetY( half( Screen->h ) - half( GetHeight() ) ); } //!< Center object vertically on the screen
        void SetXYCenter( void ) { SetXCenter(); SetYCenter(); }  //!< Center object on the screen

        bool isOnScreen() { return( ( GetX() >= 0 ) && ( GetY() >= 0 ) && ( (GetX()+GetWidth()) < Screen->w ) && ( (GetY()+GetHeight()) < Screen->h ) ); } //!< Return true if sprite is COMPLETELY visible on the screen.
        bool isPartOnScreen() { return( ( GetX() > -GetWidth() ) && ( GetY() > -GetHeight() ) && ( GetX() < Screen->w ) && ( GetY() < Screen->h ) ); } //!< Return true if sprite is PARTIALLY visible on screen.
        */

        s16 GetCenterX( void ) const { return( GetX() + half( GetWidth() ) ); } //!< Return the center's X coordinate of the object
        s16 GetCenterY( void ) const { return( GetY() + half( GetHeight() ) ); } //!< Return the center's Y coordinate of the object

        u16 Distance( const ScreenObject& Another ) const { return( pmath_2PointsDistance( GetCenterX(), GetCenterY(), Another.GetCenterX(), Another.GetCenterY() ) ); } //!< Compute the distance (in pixels) between the centers of two screen objects (member version)
        s16 Angle( const ScreenObject& Another ) const { return( pmath_2PointsAngle( GetCenterX(), GetCenterY(), Another.GetCenterX(), Another.GetCenterY() ) ); } //!< Compute the angle (in degrees) between the centers of two screen objects (member version)

        bool CollisionRect( const ScreenObject& Another, u16 Range = 0 ) const { return( pmath_collisionrect( GetCenterX(), GetCenterY(), GetWidth(), GetHeight(), Another.GetCenterX(), Another.GetCenterY(), Another.GetWidth(), Another.GetHeight(), Range ) ); } //!< Rectangular collision checker (member version)
        bool CollisionCirc( const ScreenObject& Another, u16 Range = 0 ) const { return( pmath_collisioncirc( GetCenterX(), GetCenterY(), half( GetWidth() ), Another.GetCenterX(), Another.GetCenterY(), half( Another.GetWidth() ), Range) ); } //!< Circular collision checker (member version)
        //TODO: CollisionPixel( const ScreenObject& Another, u16 Range = 0 ) ...

        //bool isMouseOver( void ) const { return( ( Mouse.X >= GetX() ) && ( Mouse.X <= GetX() + GetWidth() ) && ( Mouse.Y >= GetY() ) && ( Mouse.Y <= GetY() + GetHeight()) ); } //!< Return true if mouse is over the object
        //TODO: bool MouseClick( int flags ); // flags can be: NEW, OVER, CLICK, LEFT_CLICK, RIGHT_CLICK, MIDDLE_CLICK

};


//REMOVE?
inline void DrawScreen() { ScreenObject::DrawAll(); } //!< draw ALL registered screen objects (sprites, BGs and banners). Normally you won't invoke this function directly, WaitForVBL() will invoke it for you.

inline u16 Distance( const ScreenObject& First, const ScreenObject& Second ) { return First.Distance( Second ); } //!< Compute the distance (in pixels) between the centers of two screen objects (function version)
inline s16 Angle( const ScreenObject& First, const ScreenObject& Second ) { return First.Angle( Second ); } //!< Compute the angle (in degrees) between the centers of two screen objects (function version)

inline bool CollisionRect( const ScreenObject& First, const ScreenObject& Second, u16 Range = 0 ) { return First.CollisionRect( Second ); } //!< Rectangular collision checker (function version)
inline bool CollisionCirc( const ScreenObject& First, const ScreenObject& Second, u16 Range = 0 ) { return First.CollisionCirc( Second ); } //!< Circular collision checker (function version)
